use strict;
use warnings;
package POE::Component::Minecraft::Client::Filter;

use base 'POE::Filter';
use POE;

sub _read_keep_alive {
  my $self = shift;
  my $data = $self->_buf_read(4);
  # Packet data:
  # 0-3 uint32: Keep Alive ID

  return ["keep_alive", unpack("N", $data)];
}

sub _read_login_request {
  my $self = shift;
  my $data = $self->_buf_read(22);
  # Packet data:
  # 0-3  int32: Player Entity ID
  # 4-5  int16: string length for unused string
  # X-X  string16: UCS-2 string, unused, normally empty
  # 6-13 int64: Server map seed, needed for respawns
  #14-17 int32: Server mode, survial(0) or creative(1)
  #18-18 int8: Dimension, hell(1) or normal(0)
  #19-19 int8: Difficulty, Peaceful(0), Easy(1), Normal(2), Hard(3)
  #20-20 uint8: world height, only 128 so far
  #21-21 uint8: max players

  # this seems to be off one byte from what the wiki says...

  my ($player_ent_id, $str_len, $map_seed, $mode, $dimen, $diff, $height, $max_players) =
    unpack("N!n!qN!ccCC", $data);

  die "Login Request: String was not empty! The protocol has changed.  This library is not compatible with the server." if ($str_len != 0);
  die "Login Request: Height was not 128.  Height was ". $height.".  This library is not compatible with the server." if ($height != 128);
  
  return ["login_request",$player_ent_id, $map_seed, $mode, $dimen, $diff, $max_players];
}

sub _read_handshake {
  my $self = shift;
  my $data = $self->_buf_read(2);
  # Packet data:
  # 0-1 int16: hash string length
  # 2-N String16: hash

  my ($str_len) = unpack("n!", $data);
  die "Handshake: hash was negative length: ". $str_len if ($str_len < 0); # this could also be an unsigned, but i'm not entirely sure on it

  #I don't know if this is in characters or in bytes...

  # assume it's bytes
  my $hash = $self->_buf_read($str_len);
  warn "Handshake: buffer left is: ".$self->_buf_size()." read: ".$str_len+2;

  warn "Didn't convert from UCS-2!"; # This must be fixed! I need to decode it here, but i'm not.  I'don't know when this will be used so this should be fine for now.
  return ["handshake", $hash];
}

sub _read_chat {
  my $self = shift;
  my $data = $self->_buf_read(2);
  # Packet data:
  # 0-1 int16: hash string length (bytes or chars)
  # 2-X string16: ucs-2 encoded string

  my ($str_len) = unpack("n!", $data);
  die "Chat: message was negative length: ". $str_len if ($str_len < 0); # this could also be an unsigned, but i'm not entirely sure on it

  #I don't know if this is in characters or in bytes...

  # assume it's bytes
  my $hash = $self->_buf_read($str_len);
  warn "Chat: buffer left is: ".$self->_buf_size()." read: ".$str_len+2;

  warn "Didn't convert from UCS-2!"; # This must be fixed! I need to decode it here, but i'm not.  I'don't know when this will be used so this should be fine for now.
  return ["chat", $hash];
}

sub _read_time {
  my $self = shift;
  my $data = $self->_buf_read(8);
  # Packet data:
  # 0-7 uint64: The current game time

  my ($time) = unpack("Q", $data);

  return ["time", $time];
}

sub _read_entity_equipment {
  my $self = shift;
  my $data = $self->_buf_read(10);
  # Packet data:
  # 0-3 int32: The current game time
  # 4-5 int16: Slot on the entity
  # 6-7 int16: ID of the item
  # 8-9 int16: damage, rarely used.  Likely only for colors/types, e.g. pink wool

  my ($entity_id, $slot, $item_id, $damage) = unpack("N!n!n!n!", $data);

  return ["entity_equipment", $entity_id, $slot, $item_id, $damage];
}

sub _read_spawn_position {
  my $self = shift;
  my $data = $self->_buf_read(12);
  # Packet data:
  # 0-3  int32: X
  # 4-7  int32: y
  # 8-11 int32: z

  my ($x, $y, $z) = unpack("N!N!N!", $data);

  return ["spawn_position", $x, $y, $z];
}

sub _read_update_health {
  my $self = shift;
  my $data = $self->_buf_read(8);
  # Packet data:
  # 0-1 int16: health 0-20
  # 2-3 int16: food 0-20
  # 4-7 float: food saturation? 0.0 - 5.0? what?  no idea what this is for yet

  my ($health, $food, $saturation) = unpack("n!n!f>", $data);

  return ["update_health", $health, $food, $saturation];
}

sub _read_player_position_look {
  my $self = shift;
  my $data = $self->_buf_read(41);
  # Packet data:
  # 0-7  double: player X
  # 8-15 double: stance
  #16-23 double: player Y
  #24-31 double: player Z
  #32-35 float:  look_yaw
  #36-39 float:  look_pitch
  #40-40 bool:   on_ground

  my ($x, $s, $y, $z, $yaw, $pitch, $ground) = unpack("(FFFFff)>C", $data);

  return ["player_position_look", $x, $s, $y, $z, $yaw, $pitch, $ground];
}

# this may never be sent
sub _read_player_digging {
  my $self = shift;
  my $data = $self->_buf_read(11);
  # Packet data:
  # 0-0  int8:  action (status)
  # 1-4  int32: block X
  # 5-5  uint8: block Y
  # 6-9  int32: block Z
  #10-10 int8:  face

  my ($action, $x, $y, $z, $face) = unpack("cN!CN!c", $data);

  return ["player_digging", $action, $x, $y, $z, $face];
}

sub _read_player_block_placement {
  my $self = shift;
  my $data = $self->_buf_read(12);
  # Packet data:
  # 0-3  int32: block X
  # 4-4  uint8: block Y
  # 5-8  int32: block Z
  # 9-9  int8:  direction
  #10-11 int16: block/item id

  # optional data: if (block/item id > 0)
  #12-12 int8:  amount
  #13-14 int16: damage/data

  my ($x, $y, $z, $dir, $block_id) = unpack("N!CN!cn!", $data);
  
  if ($block_id > 0) {
    my $data2 = $self->_buf_read(3);
    my ($amount, $damage) = unpack("cn!", $data2);
    return ["player_block_placement", $x, $y, $z, $dir, $block_id, $amount, $damage];
  } else {
    return ["player_block_placement", $x, $y, $z, $dir, $block_id];
  }
}

sub _read_player_use_bed {
  my $self = shift;
  my $data = $self->_buf_read(14);
  # Packet data:
  # 0-3  int32: entity_id
  # 4-4  int8:  in bed?
  # 5-8  int32: block X
  # 9-9  uint8: block Y
  #10-13  int32: block Z

  # in bed, seems to always be zero according to the docs, gonna setup an assert on it for testing

  my ($entity_id, $in_bed, $x, $y, $z) = unpack("N!cN!CN!", $data);
  return ["player_use_bed", $entity_id, $in_bed, $x, $y, $z);
}

sub _read_animation {
  my $self = shift;
  my $data = $self->_buf_read(5);
  # Packet data:
  # 0-3  int32: entity_id
  # 4-4  int8:  animation

  my ($entity_id, $animation) = unpack("N!c", $data);
  return ["animation", $entity_id, $animation);
}

sub _read_named_entity_spawn {
  my $data = $self->_buf_read(6);
  # Packet data:
  # 0-3  int32: entity_id
  # 4-X0 string16:  Entity Name
  #X1-X4 int32:  X
  #X5-X5 int8:   Y
  #X6-X10 int32: Z
  #X11-X11 uint8: Rotation ($_ / 256 * 360)
  #X12-X12 uint8: Pitch ($_ / 256 * 360)
  #X13-X14 int16: item held

  my ($entity_id, $str_len) = unpack("N!n!", $data);
  die "Named entity spawn: hash was negative length: ". $str_len if ($str_len < 0); # this could also be an unsigned, but i'm not entirely sure on it

  #I don't know if this is in characters or in bytes...

  # assume it's bytes
  my $name = $self->_buf_read($str_len);
  warn "Named entity spawn: buffer left is: ".$self->_buf_size()." read: ".$str_len+2;

  warn "Didn't convert from UCS-2!"; # This must be fixed! I need to decode it here, but i'm not.  I'don't know when this will be used so this should be fine for now.

  $data = $self->buf_read(14);

  my ($x, $y, $z, $rotation, $pitch, $item) = unpack("N!CN!CCn!", $data);
  return ["Named Entity Spawn", $entity_id, $name, $x, $y, $z, $rotation, $pitch, $item);
}

sub _read_pickup_spawn {
  my $data = $self->_buf_read(24);
  # Packet data:
  # 0-3  int32: entity_id
  # 4-5  int16: item_id
  # 6-6  int8:  count
  # 7-8  int16: damage
  # 9-12 int32: X
  #13-16 int32: Y  # why use an int32 for Y here, but not elsewhere?
  #17-20 int32: Z
  #21-21 int8:  Rotation # OMG! not using quaternions!
  #22-22 int8:  Pitch
  #23-23 int8:  Roll

  my ($entity_id, $item_id, $count, $damage, $x, $y, $z, $rotation, $pitch, $roll) = unpack("N!n!Cn!N!CN!CCC", $data);
  return ["pickup_spawn", $entity_id, $item_id, $count, $damage, $x, $y, $z, $rotation, $pitch, $roll);
}

sub _read_collect_item {
  my $data = $self->_buf_read(8);
  # Packet data:
  # 0-3  int32: pickup entity_id
  # 4-7  int32: collector entity_id
  
  my ($entity_id, $collector_id) = unpack("N!N!", $data);
  return ["collect_item", $entity_id, $collector_id);
}

sub _read_add_object {
  my $data = $self->_buf_read(8);
  # Packet data:
  # 0-3  int32: entity_id
  # 4-4  int8: type of object
  # 5-8  int32: X
  # 9-12 int32: Y
  #13-16 int32: Z
  #17-20 int32: Fireball thrower's ID

  #Optional Data: if (fireball_id != 0)
  #21-22 int16: FB X
  #23-24 int16: FB Y
  #25-26 int16: FB Z  

  my ($entity_id, $type, $x, $y, $z, $thrower_id) = unpack("N!CN!N!N!N!", $data);
  my ($fb_x, $fb_y, $fb_z);

  if ($thrower_id) { # is negative something else? dunno
    my $data = $self->_buf_read(6);
    ($fb_x, $fb_y, $fb_z) = unpack("n!n!n!", $data);
  }

  return ["add_object", $entity_id, $type, $x, $y, $z, $thrower_id, $fb_x, $fb_y, $fb_z];
}

my $read_dispatch = {
  0x00 => \&_read_keep_alive,
  0x01 => \&_read_login_request,
  0x02 => \&_read_handshake,
  0x03 => \&_read_chat,
  0x04 => \&_read_time,
  0x05 => \&_read_entity_equipment,
  0x06 => \&_read_spawn_position,
  0x08 => \&_read_update_health,
  0x0D => \&_read_player_position_look,
  0x0E => \&_read_player_digging, #CTS?
  0x0F => \&_read_player_block_placement, #CTS?
#  0x10 => \&_read_player_holding, #CTS
  0x11 => \&_read_player_use_bed,
  0x12 => \&_read_animation, #CTS?
# 0x13 => CTS only, only needed for writing
  0x14 => \&_read_named_entity_spawn,
  0x15 => \&_read_pickup_spawn,
  0x16 => \&_read_collect_item, # only used for animation, other packets are for inventory and deletion of ids
  0x17 => \&_read_add_object, # /vehicle
  0x18 => \&_read_mob_spawn,
  0x19 => \&_read_painting, # tells us what the painting is
  0x1A => \&_read_experience_orb,
# 0x1B => Unused
  0x1C => \&_read_entity_velocity,
  0x1D => \&_read_destroy_entity,
  0x1E => \&_read_entity,
  0x1F => \&_read_entity_relative_move,
  0x20 => \&_read_entity_look, # / rotate
  0x21 => \&_read_entity_look_relative_move,
  0x22 => \&_read_entity_teleport,
# 0x23 => Unused
# 0x24 => Unused
# 0x25 => Unused
  0x26 => \&_read_status,
  0x27 => \&_read_attach_entity,
  0x28 => \&_read_entity_metadata,
  0x29 => \&_read_entity_effect,
  0x2A => \&_read_remove_entity_effect,
# 0x2B => Unused
# ...
# 0x31 => Unused
  0x32 => \&_read_pre_chunk,
  0x33 => \&_read_map_chunk,
  0x34 => \&_read_multi_block_change,
  0x35 => \&_read_block_change,
  0x36 => \&_read_block_action,
# 0x37 => Unused
# ...
# 0x3B => Unused
  0x3C => \&_read_explosion,
  0x3D => \&_read_sound_effect,
# 0x3E => Unused
# ...
# 0x45 => Unused
  0x46 => \&_read_new_invalid_state,
  0x47 => \&_read_thunderbolt,
# 0x48 => Unused
# ...
# 0x63 => Unused
  0x64 => \&_read_open_window,
  0x65 => \&_read_close_window,
  0x66 => \&_read_window_click,
  0x67 => \&_read_set_slot,
  0x68 => \&_read_window_items,
  0x69 => \&_read_progress_bar,
  0x6A => \&_read_transaction,
  0x6B => \&_read_creative_inventory_action, # CTS?
# 0x6C => Unused,
# ...
# 0x81 => Unused,
  0x82 => \&_read_update_sign,
  0x83 => \&_read_item_data,
# 0x84 => Unused,
# ...
# 0xC7 => Unused,
  0xC8 => \&_read_increment_statistic,
  0xC9 => \&_read_player_list_item,
# 0xCA => Unused,
# ...
# 0xFD => Unused,
# 0xFE => \&_read_server_list_ping,
  0xFF => \&_read_disconnect,
};

my $write_dispatch = {
  keep_alive                => \&_write_keep_alive,
  login_request             => \&_write_login_request,
  handshake                 => \&_write_handshake,
  chat                      => \&_write_chat,
  entity_equipment          => \&_write_entity_equipment,
  use_entity                => \&_write_use_entity,
  respawn                   => \&_write_respawn,
  player                    => \&_write_player, # i want to give this a new name, on_ground?
  player_position           => \&_write_player_position,
  player_look               => \&_write_player_look,
  player_position_look      => \&_write_player_position_look,
  player_digging            => \&_write_player_digging,
  player_block_placement    => \&_write_player_block_placement,
  holding_change            => \&_write_holding_change,
  animation                 => \&_write_animation, # STC?
  entity_action             => \&_write_entity_action,
  pickup_spawn              => \&_write_pickup_spawn,
  painting                  => \&_write_painting, # STC?
# stance_update             => \&_write_stance_update,
  entity_velocity           => \&_write_entity_velocity,
  attach_entity             => \&_write_attach_entity,
# multi_block_change        => \&_write_multi_block_change, may need this don't know
  block_change              => \&_write_block_change,
  window_click              => \&_write_window_click,
  transaction               => \&_write_transaction,
  creative_inventory_action => \&_write_creative_inventory_action, # STC?
  update_sign               => \&_write_update_sign,
  increment_statistic       => \&_write_increment_statistic,
  server_list_ping          => \&_write_server_list_ping,
};

sub new {
  my $class = shift;

  my $self = bless [''], $class;

  return $self;
}

sub get_one_start {
  my $self = shift;
  my $packets = shift;

  $self->[0] .= $_ for (@$packets);
}

sub get_one {
  die "Not implemented yet";
}

sub put {
  die "Not implemented yet";
}

sub get_pending {
  my $self = shift;

  return [$self->[0]] if length $self->[0];
  return undef;
}